@@ -25,6 +25,10 @@
 
 #include <gnome.h>
 #include <pwd.h>
+#include <libgnomevfs/gnome-vfs.h>
+#include <libgnomevfs/gnome-vfs-mime-handlers.h>
+#include <libgnomevfs/gnome-vfs-mime-utils.h>
+#include <libgnomeui/gnome-thumbnail.h>
 #include <gconf/gconf-client.h>
 #include <glade/glade.h>
 #include <pwd.h>
@@ -47,6 +51,7 @@ typedef struct {
 
 	GdkScreen    	*screen;
 	GtkIconTheme 	*theme;
+	GnomeThumbnailFactory *thumbs;
 
 	EContactAddress *addr1;
 	EContactAddress *addr2;
@@ -208,6 +213,9 @@ about_me_focus_out (GtkWidget *widget, GdkEventFocus *event, GnomeAboutMe *me)
 	gint i;
 	
 	wid = glade_get_widget_name (widget);
+
+	if (wid == NULL)
+		return FALSE;
 	
 	for (i = 0; ids[i].wid != NULL; i++)
 		if (g_ascii_strcasecmp (ids[i].wid, wid) == 0)
@@ -262,8 +270,13 @@ get_user_login (void)
 	struct passwd pwd, *err;
 
 	int i;
-	i = getpwuid_r(getuid(), &pwd, buf, sizeof(buf), &err);
-	return ((i == 0) && (err == &pwd)) ? g_strdup(pwd.pw_name) : NULL;
+#if __sun
+	i = getpwuid_r (getuid (), &pwd, buf, sizeof (buf));
+	return (i != 0) ? g_strdup (pwd.pw_name) : NULL;
+#else
+	i = getpwuid_r (getuid (), &pwd, buf, sizeof (buf), &err);
+	return ((i == 0) && (err == &pwd)) ? g_strdup (pwd.pw_name) : NULL;
+#endif
 }
 
 /*
@@ -404,9 +417,9 @@ about_me_load_photo (GnomeAboutMe *me, EContact *contact)
 
 	photo = e_contact_get (contact, E_CONTACT_PHOTO);
 
-	if (photo) {
+	if (photo && photo->type == E_CONTACT_PHOTO_TYPE_INLINED) {
 		me->have_image = TRUE;
-		e_image_chooser_set_image_data (E_IMAGE_CHOOSER (widget), photo->data, photo->length);
+		e_image_chooser_set_image_data (E_IMAGE_CHOOSER (widget), photo->data.inlined.data, photo->data.inlined.length);
 		e_contact_photo_free (photo);
 	} else {
 		me->have_image = FALSE;
@@ -477,15 +490,16 @@ about_me_update_photo (GnomeAboutMe *me)
 		}
 
 		photo = g_new0 (EContactPhoto, 1);
-		photo->data = data;
-		photo->length = length;
+		photo->type = E_CONTACT_PHOTO_TYPE_INLINED;
+		photo->data.inlined.data = data;
+		photo->data.inlined.length = length;
 		e_contact_set (me->contact, E_CONTACT_PHOTO, photo);
 
 		/* Save the image for GDM */
 		/* FIXME: I would have to read the default used by the gdmgreeter program */
 		file = g_strdup_printf ("%s/.face", g_get_home_dir ());
 		fp = fopen (file, "wb");
-		fwrite (photo->data, 1, photo->length, fp);
+		fwrite (photo->data.inlined.data, 1, photo->data.inlined.length, fp);
 		fclose (fp);
 
 		g_free (file);	
@@ -540,6 +554,42 @@ eab_create_image_chooser_widget (gchar *name,
 	return w;
 }
 
+static void 
+about_me_update_preview (GtkFileChooser *chooser,
+			 GnomeAboutMe   *me) 
+{
+	gchar *uri;
+
+	uri = gtk_file_chooser_get_preview_uri (chooser);
+	
+	if (uri) {
+		GtkWidget *image;
+		GdkPixbuf *pixbuf;
+		gchar *mime_type;
+
+		if (!me->thumbs)
+			me->thumbs = gnome_thumbnail_factory_new (GNOME_THUMBNAIL_SIZE_NORMAL);
+
+		
+		mime_type = gnome_vfs_get_mime_type (uri);
+		pixbuf = gnome_thumbnail_factory_generate_thumbnail (me->thumbs,
+								     uri,
+								     mime_type);
+		image = gtk_file_chooser_get_preview_widget (chooser);
+		
+		if(pixbuf != NULL) {
+			gtk_image_set_from_pixbuf (GTK_IMAGE (image), pixbuf);
+			g_object_unref (pixbuf);
+		} else {
+			gtk_image_set_from_stock (GTK_IMAGE (image),
+						  "gtk-dialog-question",
+						  GTK_ICON_SIZE_DIALOG);
+		}
+		g_free (mime_type);
+	}
+	gtk_file_chooser_set_preview_widget_active (chooser, TRUE);
+}
+
 static void
 about_me_image_clicked_cb (GtkWidget *button, GnomeAboutMe *me)
 {
@@ -547,6 +597,7 @@ about_me_image_clicked_cb (GtkWidget *button, GnomeAboutMe *me)
 	gint response;
 	GtkWidget *image_chooser;
 	GladeXML  *dialog;
+	GtkWidget *image;
 	
 	dialog = me->dialog;
 	image_chooser = WID ("image-chooser");
@@ -561,6 +612,19 @@ about_me_image_clicked_cb (GtkWidget *button, GnomeAboutMe *me)
 	gtk_dialog_set_default_response (GTK_DIALOG (chooser_dialog), GTK_RESPONSE_ACCEPT);
 	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (chooser_dialog), g_get_home_dir ());
 
+	gtk_file_chooser_set_use_preview_label (GTK_FILE_CHOOSER (chooser_dialog),
+						FALSE);
+	
+	image = gtk_image_new ();
+	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER (chooser_dialog),
+					     image);
+	gtk_widget_set_size_request (image, 128, -1);
+	
+	gtk_widget_show (image);
+
+	g_signal_connect (chooser_dialog, "update-preview",
+			  G_CALLBACK (about_me_update_preview), me);
+
 	response=gtk_dialog_run (GTK_DIALOG (chooser_dialog));
 
 	if (response == GTK_RESPONSE_ACCEPT) {
